import pytest
from marshmallow import ValidationError

from mckinseysolvegame.domain.models import OptimizationResult, Species


# region deserialization
def test_deserialize_species():
    # Arrange
    species_json = '[{"name": "Species 1", "caloriesProvided": 2, \
        "caloriesNeeded": 3, "depthRange": "", "temperatureRange": "", "foodSources": []}, \
        {"name": "Species 2", "caloriesProvided": 4, "caloriesNeeded": 2, "depthRange": "", "temperatureRange": "", \
            "foodSources": []}]'

    # Act
    species = Species.from_json(species_json, many=True)

    # Assert
    assert len(species) == 2
    assert isinstance(species[0], Species)
    assert species[0].name == 'Species 1'
    assert species[0].calories_provided == 2
    assert species[0].calories_needed == 3
    assert species[0].depth_range == ""
    assert species[0].temperature_range == ""
    assert species[0].food_sources == []
    assert isinstance(species[1], Species)
    assert species[1].name == 'Species 2'
    assert species[1].calories_provided == 4
    assert species[1].calories_needed == 2
    assert species[0].depth_range == ""
    assert species[0].temperature_range == ""
    assert species[1].food_sources == []


def test_deserialize_species_with_long_name():
    # Arrange
    species_json = '{"name": "A very long contract name that exceeds the \
        maximum length of 64 characters", ' \
        '"caloriesProvided": 2, "caloriesNeeded": 3, "depthRange": "", "temperatureRange": "", "foodSources": []}'

    # Act
    with pytest.raises(ValidationError) as e:
        _ = Species.from_json(species_json)

    # Assert
    assert 'name' in e.value.messages
    assert 'Longer than maximum length 64.' in e.value.messages['name']


def test_deserialize_species_with_negative_calories_provided():
    # Arrange
    species_json = '{"name": "Species 1", "caloriesProvided": -2, \
        "caloriesNeeded": 3, "depthRange": "", "temperatureRange": "", "foodSources": []}'
    # Act
    with pytest.raises(ValidationError) as e:
        _ = Species.from_json(species_json)
    # Assert
    assert 'caloriesProvided' in e.value.messages
    assert 'Must be greater than or equal to 1.' \
        in e.value.messages['caloriesProvided']


def test_deserialize_species_with_negative_calories_needed():
    # Arrange
    species_json = '{"name": "Species 1", "caloriesProvided": 1, \
        "caloriesNeeded": -3, "depthRange": "", "temperatureRange": "", "foodSources": []}'
    # Act
    with pytest.raises(ValidationError) as e:
        _ = Species.from_json(species_json)
    # Assert
    assert 'caloriesNeeded' in e.value.messages
    assert 'Must be greater than or equal to 0.' \
        in e.value.messages['caloriesNeeded']


def test_deserialize_species_with_wrong_type():
    invalid_species = {"name": "invalid", "caloriesProvided": "not_an_int",
                       "caloriesNeeded": 10, "depthRange": "", "temperatureRange": "", "foodSources": []}
    with pytest.raises(ValidationError):
        _ = Species.from_json(invalid_species)

    invalid_species = {"name": "invalid", "caloriesProvided": 10,
                       "caloriesNeeded": "not_an_int", "depthRange": "", "temperatureRange": "", "foodSources": []}
    with pytest.raises(ValidationError):
        _ = Species.from_json(invalid_species)

    invalid_species = {"name": "invalid", "caloriesProvided": 10,
                       "caloriesNeeded": 10, "depthRange": "", "temperatureRange": "", "foodSources": [1, ]}
    with pytest.raises(ValidationError):
        _ = Species.from_json(invalid_species)


def test_deserialize_species_with_missing_required_field():
    json_data = '{"name": "Species 1", "caloriesProvided": 2, \
        "caloriesNeeded": 3, "depthRange": "", "temperatureRange": ""}'
    with pytest.raises(ValidationError) as e:
        _ = Species.from_json(json_data)
    assert "foodSources" in e.value.messages
    assert "required field" in str(e.value)


def test_deserialize_species_with_extra_field():
    json_data = '{"name": "Species 1", "caloriesProvided": 2, \
        "caloriesNeeded": 3, "depthRange": "", "temperatureRange": "", "foodSources": [], "extraField": "value"}'
    with pytest.raises(ValidationError) as error:
        _ = Species.from_json(json_data)
    assert "extraField" in error.value.messages
    assert "Unknown field." in str(error.value)


def test_optimization_result_deserialization():
    # Test valid deserialization
    input_data = {
        'species': ['Species1', 'Species2', 'Species3']
    }
    try:
        result = OptimizationResult.from_json(input_data)
        assert isinstance(result, OptimizationResult)
        assert result.species == ['Species1', 'Species2', 'Species3']
    except ValidationError:
        pytest.fail('Deserialization should not raise an error.')

    # Test deserialization with missing required fields
    invalid_data = [
        {},
    ]
    for data in invalid_data:
        with pytest.raises(ValidationError):
            _ = OptimizationResult.from_json(data)

    # Test deserialization with invalid field types
    invalid_data = [
        {'species': 'Contract1'},
        {'species': ['Contract1', 2, 'Contract3']},
    ]
    for data in invalid_data:
        with pytest.raises(ValidationError):
            _ = OptimizationResult.from_json(data)

# endregion


def test_serialize_species():
    species = Species("Species 1", 2, 3, "", "", [])
    serialized = species.to_json("dict")
    expected = {
        "name": "Species 1",
        "caloriesProvided": 2,
        "caloriesNeeded": 3,
        "depthRange": "",
        "temperatureRange": "",
        "foodSources": []
    }
    assert serialized == expected


def test_optimization_result_serialization():
    # Test valid serialization
    result = OptimizationResult(species=['Species1', 'Species2', 'Species3'])
    try:
        serialized = result.to_json()
        assert isinstance(serialized, dict)
        assert serialized == {'species': ['Species1', 'Species2', 'Species3']}
    except ValidationError:
        pytest.fail('Serialization should not raise an error.')



import copy
from itertools import combinations, groupby
from typing import List

from mckinseysolvegame.domain.models import Species


class Solver:

    @staticmethod
    def find_sustainable_food_chain(species: List[Species]) -> dict:
        maximum_food_chain_length = 8
        species_copy = copy.deepcopy(species)
        if not species_copy:
            return {}

        populate_food_sources(species_copy)

        species_copy.sort(key=lambda x: x.depth_range)
        grouped_species = {key: list(group) for key, group in groupby(
            species_copy, key=lambda x: x.depth_range)}

        longest_sustainable_chain_per_depth_range = {}
        for depth_range, species_copy in grouped_species.items():
            n = min(maximum_food_chain_length, len(species_copy))
            species_copy.sort(key=lambda x: x.calories_provided, reverse=True)

            optimal_list = []
            for length in range(1, n + 1):
                for combination in combinations(species_copy, length):
                    combination = list(combination)
                    if is_sustainable(combination) and len(combination) > len(optimal_list):
                        optimal_list = combination

            optimal_list.sort(key=lambda x: x.calories_provided, reverse=True)
            eating_steps = simulate_eating(optimal_list)
            longest_sustainable_chain_per_depth_range[depth_range] = eating_steps

        _, max_value = max(
            longest_sustainable_chain_per_depth_range.items(), key=lambda x: len(x[1]))

        return max_value


def populate_food_sources(species: List[Species]) -> None:
    for s in species:
        if not s.food_sources:
            continue
        food_sources_as_species = []
        for food_source in s.food_sources:
            matching_species = [x for x in species if x.name == food_source]
            if matching_species:
                food_sources_as_species.append(matching_species[0])
        s.food_sources = food_sources_as_species


def is_sustainable(species: List[Species]) -> bool:
    if not species:
        return False

    if not any(s.calories_needed == 0 for s in species):
        return False

    species_dict = simulate_eating(species)

    for value in list(species_dict.values()):
        if value['calories_needed'] != 0 or value['calories_provided'] <= 0:
            return False

    return True


def simulate_eating(species: List[Species]):
    species_dict = {s.name: {'calories_needed': s.calories_needed,
                             'calories_provided': s.calories_provided} for s in species}

    for s in sorted(species, key=lambda x: x.calories_provided, reverse=True):
        if s.calories_needed == 0 or not s.food_sources:
            # does not need to eat (producer) or cannot eat
            continue

        s.food_sources.sort(key=lambda x: x.calories_provided, reverse=True)

        if len(s.food_sources) > 1 and \
                s.food_sources[0].name in species_dict and \
                s.food_sources[1].name in species_dict and \
                species_dict[s.food_sources[1].name]['calories_provided'] == species_dict[s.food_sources[0].name]['calories_provided'] and \
                species_dict[s.food_sources[0].name]['calories_provided'] >= species_dict[s.name]['calories_needed'] / 2:
            half_calories_needed = int(species_dict[s.name]['calories_needed'] / 2)
            species_dict[s.food_sources[0].name]['calories_provided'] -= half_calories_needed
            species_dict[s.food_sources[1].name]['calories_provided'] -= half_calories_needed
            species_dict[s.name]['calories_needed'] = 0
            species_dict[s.name]['eats'] = [s.food_sources[0].name, s.food_sources[1].name]
            continue  # only eat once

        for food in s.food_sources:
            if food.name in species_dict and species_dict[food.name]['calories_provided'] > species_dict[s.name]['calories_needed']:
                species_dict[food.name]['calories_provided'] -= species_dict[s.name]['calories_needed']
                species_dict[s.name]['calories_needed'] = 0
                species_dict[s.name]['eats'] = [food.name]
                break  # only eat once

    return species_dict
